#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import base64
import struct

from ctypes import cdll, c_char_p, c_void_p, c_ulonglong

class QuickXORHash:
    def __init__(self):
        try:
            self.libqxh = cdll.LoadLibrary('libqxh.so.0')
        except OSError:
            self.HAS_LIBQXH = False
        else:
            self.HAS_LIBQXH = True
            self.libqxh.qxh_new.restype = c_void_p
            self.libqxh.qxh_update.argtypes = [c_void_p, c_char_p, c_ulonglong]
            self.libqxh.qxh_finalize.argtypes = [c_void_p]
            self.libqxh.qxh_finalize.restype = c_char_p
            self.libqxh.qxh_free.argtypes = [c_void_p]
            self.qxh = self.libqxh.qxh_new()
            return

        # Constants
        self.width = 160
        self.shift = 11

        # State
        self.shifted = 0
        self.length = 0
        self.cell = [0] * (int((self.width - 1) / 64) + 1)

    def update(self, data):
        if self.HAS_LIBQXH:
            self.libqxh.qxh_update(self.qxh, str(data), len(data))
            return

        cell_index = int(self.shifted / 64)
        cell_bitpos = int(self.shifted % 64)

        for i in range(0, min(self.width, len(data))):
            next_cell = cell_index + 1
            cell_bits = 64
            # Last cell needs to wrap around
            if next_cell == len(self.cell):
                next_cell = 0
                # Last cell usually isn't a full 64 bits
                if self.width % 64 > 0:
                    cell_bits = self.width % 64

            new_byte = 0
            for j in range(i, len(data), self.width):
                new_byte ^= data[j]

            # Python doesn't have fixed-width data types, so we need to
            # explicitly throw away extra bits.
            self.cell[cell_index] ^= new_byte << cell_bitpos & 0xffffffffffffffff

            if cell_bitpos > cell_bits - 8:
                self.cell[next_cell] ^= new_byte >> (cell_bits - cell_bitpos)

            cell_bitpos += self.shift
            if cell_bitpos >= cell_bits:
                cell_index = next_cell
                cell_bitpos -= cell_bits

        self.shifted += self.shift * (len(data) % self.width)
        self.shifted %= self.width
        self.length += len(data)

    def finalize(self):
        if self.HAS_LIBQXH:
            digest = self.libqxh.qxh_finalize(self.qxh)
            self.libqxh.qxh_free(self.qxh)
            return digest

        # Convert cells to byte array
        b_data = bytearray()
        for i in range(0, len(self.cell)):
            chunk = struct.unpack('8B', struct.pack('Q', self.cell[i]))
            if (i + 1) * 64 <= self.width:
                b_data.extend(chunk)
            else:
                b_data.extend(chunk[0:int(self.width / 8 % 8)])

        # Convert length to byte array
        b_length = struct.unpack('8B', struct.pack('Q', self.length))

        # XOR the length with the least significant bits
        offset = int((self.width / 8) - len(b_length))
        for i in range(0, len(b_length)):
            b_data[i + offset] ^= b_length[i]

        return base64.b64encode(b_data)

    def hash_file(self, path):
        with open(path, 'rb') as f:
            while True:
                chunk = f.read(512 * 1024)
                if chunk:
                    self.update(bytearray(chunk))
                else:
                    break
        return self.finalize()



#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json
import sys

import odm.cli

def main():
    odm.cli.CLI.writer_wrap(sys)
    cli = odm.cli.CLI(['action'])
    client = cli.client

    if cli.args.action == 'list-users':
        print(json.dumps(client.list_users(), indent = 2))

    else:
        print('Unsupported action {}'.format(cli.args.action), file = sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import argparse
import logging
import sys

import yaml

from kitchen.text.converters import getwriter

from odm import googledriveclient, onedriveclient

class CLI:
    def __init__(self, args, client='microsoft'):
        parser = argparse.ArgumentParser()
        parser.add_argument('-c', '--config',
            help = 'Config file location',
            default = '/etc/odm.yaml',
        )
        parser.add_argument('-v', '--verbose',
            help = 'Enable verbose output',
            action = 'store_true',
        )
        for arg in args:
            parser.add_argument(arg)

        self.args = parser.parse_args()

        with open(self.args.config, 'r') as configfile:
            self.config = yaml.safe_load(configfile)

        self.config['args'] = self.args

        self.logger = logging.getLogger(__name__)
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(logging.Formatter('%(asctime)s %(message)s', '%Y-%m-%dT%H:%M:%S'))
        if self.args.verbose:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        self.logger.addHandler(handler)

        if client == 'google':
            self.client = googledriveclient.GoogleDriveClient(self.config, self.logger)
        elif client == 'microsoft':
            self.client = onedriveclient.OneDriveClient(self.config, self.logger)

    @staticmethod
    def writer_wrap(caller_sys):
        writer = getwriter('utf8')
        caller_sys.stdout = writer(caller_sys.stdout)
        caller_sys.stderr = writer(caller_sys.stderr)



#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import base64
import json
import os
import random
import re
import requests
import sys
import time

import requests
import requests_oauthlib
import requests_toolbelt

from bs4 import BeautifulSoup
from oauthlib.oauth2 import BackendApplicationClient

from odm import inkml, quickxorhash

KETSUBAN = '''
iVBORw0KGgoAAAANSUhEUgAAAMkAAADhCAYAAABiOZFeAAAFVElEQVR42u3dvW1bMRSAUffuvGNG
SJsRPIen8wpKKzzAFHl1L3+s8wGqZMUIxFPQ5CPf3iRpZh/v77f7V29//3z++Gr97DP/TuRzvmFB
AokggUSLuwVrDdLoYK8A5BsWJJAIEkgECSR6DSQZg7sCrG9YkEAiSCDRoUgq5g9Z8xVIBAkkggQS
QQKJzkUysv6RMSHPQNLKNyxIIBEkkGhx31//bvev6vnJyPpKxrqMb1iQQCJIIBEkkGjvro/v3oOp
eC/rEV2P7woSSAQJJDqk6wC7zlGyX9G5BSSCBBJBAokggUS/s4p1korHfqP7xnzDggQSQQKJFtda
07i+Wj/X+14UVPT3+YYFCSSCBBJBAon2bvZpixV/AHBgtiCBRJBAooPmJNVPEc44nM46iSCBRJBA
Ikgg0bkT9+oB3VrTiL6skwgSSAQJJNq4rG3tFVveo/Mj6ySCBBJBAokggUTnIsmYVLe2249AyPj9
vmFBAokggUSLm3E4Xe+aRnSLjK3yggQSQQKJIIFE59aaWK9eJ4msmTjBUZBAIkgg0eZIIvOMR2sj
vXOQ6M9aJxEkkAgSSAQJJDq36wDLuIxn5NFe97gLEkgECSR6MSTRE1Gigzvjc9ZJBAkkggQSQQKJ
zi1rW3vv2kjVffBOSxEkkAgSSLRpFdvTK+5h9GSiIIFEkEAiSCDRa0zcKw63zj6VceQPCr5hQQKJ
IIFEG89Jsu4uqb6G2pxEkEAiSCARJJAor+gjqzPuKawY7Bn/pq3ykEACiSCBRBPmCFnbRKJb5Xtf
WddQu6IaEkggESSQCBJIVFP0RMUokqw9Ur0HWGed/Nj7nhEFCSSQQAIJJFowJ/lpDlIB5vqzkAgS
SAQJJIIEEkHS+4eBilNPZqyT2LsFCSSQQAIJJGoUPb1k9Tb66kPmzEkECSSCBBJBAonmVLF3a2SA
RddNetdQsu5M9PguJJBAIkggUdI6ScbTgFl3GFafbFJxcJ0RBQkkkEACCSSCBBKtmbhHB2YGyqzT
UqITd5f4QAIJJJBAAokGkPSeQhJ9qrA1+FpbZLK2ntiWIkggESSQCBJItPc6yewTHKOnpfTu65q9
/d6IggQSSCCBBBI9mJNUX++8+qnFiu0skEACCSSQQAKJIIFE8Xa6qz3rEdrq39f6Y4QRBQkkkEAC
CSS6NONwuvtaaxoVSCrWflprNkYUJJBAAgkkkAgSSDTWdTBkbJUfeS96AkvGvq7oq/WIsBEFCSSQ
QAIJJHowJ8m4NzC6LSX6ZGJr/tAa4NH3rJNAAgkkggQSQQKJzkUyAqF38hx9fDf6nnUSSCCBRJBA
oifWSSruBMnY4h79XPWayRW6EQUJJJBAAgkkggQSPYekupFLgzLWSWa8TNwhgQQSSCCBRBshmXEo
d/UdJO4ngQQSSAQJJIIEEr0GkurTJB2YLUggESSQaEKz1xSyBnQ1tOgjBEYUJJBAAgkkkAgSSLT3
xL3iXsToAd0jJ0haJ4EEEkgECST6hUgq5ha9B9BFt/QbUZBAAgkkkEAiSCDRHCSz719fuf9r5P9k
REECCSSQQAKJLu20Bb3i9824T9H9JJBAAgkkkEAiSCDRfCQVdyZmbZXPgB59GVGQQAIJJJBAoqR1
kpFBGh2IGZ/LWkPxZCIkkEAiSCARJJBo74l79YQ4Wtbju63PucQHEkgggQQSSFSAZMY29+g8Z+X2
eyMKEkgggQQSSAQJJFozce8dVLNPXanYfm+dBBJIIBEkkGigUw6Zq/j9FWCNKEgggQQSSCCRJEmS
JEmSJEmSJEmSJEmSJEmSJEmSJEmSJEmSJEmSJEmSjuo/MassmDD1NGYAAAAASUVORK5CYII=
'''

class OneDriveClient:
    def __init__(self, config, logger):
        self.baseurl = 'https://graph.microsoft.com/v1.0/'
        self.config = config
        self.logger = logger
        client = BackendApplicationClient(client_id = config['microsoft']['client_id'])
        self.msgraph = requests_oauthlib.OAuth2Session(client = client)
        self._get_token()

    def _get_token(self):
        self.token = self.msgraph.fetch_token(
            token_url = 'https://login.microsoftonline.com/{}/oauth2/v2.0/token'.format(self.config.get('domain')),
            client_id = self.config['microsoft']['client_id'],
            client_secret = self.config['microsoft']['client_secret'],
            scope = [ 'https://graph.microsoft.com/.default' ],
        )

    def _get(self, path):
        if self.token['expires_at'] < time.time() + 300:
            self._get_token()

        if self.baseurl not in path:
            path = ''.join([self.baseurl, path])

        return self.msgraph.get(
            path,
            timeout = self.config.get('timeout', 60),
            allow_redirects = False,
        )

    def get(self, path):
        result = None
        page_result = None
        attempt = 0

        while not page_result:
            attempt += 1
            error = None
            delay = random.uniform(0, min(300, 3 * 2 ** attempt))
            try:
                page_result = self._get(path)
            except (
                requests.exceptions.ReadTimeout,
                requests.exceptions.ConnectionError,
            ) as e:
                self.logger.warn(e)
                error = 'requests error'

            if page_result is not None:
                if page_result.status_code == 429:
                    if 'retry-after' in page_result.headers:
                        delay = page_result.headers['retry-after']
                    error = 'Throttled'
                elif page_result.status_code == 504:
                    error = 'Gateway timeout'

            if error:
                self.logger.warn('{}, sleeping for {} seconds'.format(error, delay))
                time.sleep(delay)
                page_result = None
                continue

            if page_result.status_code == 302:
                return { 'location': page_result.headers['location'] }
            elif page_result.status_code == 404:
                return None
            else:
                page_result.raise_for_status()
                decoded = json.loads(page_result.content)
                if result:
                    result['value'].extend(decoded['value'])
                else:
                    result = decoded
                if '@odata.nextLink' in decoded:
                    self.logger.debug('Getting next page...')
                    path = decoded['@odata.nextLink']
                    page_result = None
                    attempt = 0

        return result

    def list_users(self):
        users = self.get('users?$select=id,displayName,givenName,jobTitle,mail,userPrincipalName,accountEnabled,onPremisesImmutableId,onPremisesSyncEnabled')
        if users:
            return users['value']
        return []

    def show_user(self, user):
        return self.get('/users/{}@{}'.format(user, self.config['domain']))

    def list_drives(self, user):
        drives = self.get('users/{}@{}/drives'.format(user, self.config['domain']))
        if drives:
            for d in drives['value']:
                d['root'] = self.get('drives/{}/root'.format(d['id']))
            return drives['value']
        return []

    def list_folder(self, folder):
        return self.get('drives/{}/items/{}/children?select=file,folder,id,name,package,parentReference,remoteItem,size,fileSystemInfo,malware'.format(folder['parentReference']['driveId'], folder['id']))['value']

    def expand_items(self, items):
        expanded = True
        while expanded:
            expanded = False
            for item in items:
                if ('folder' in item or 'package' in item) and 'expanded' not in item:
                    items.extend(self.list_folder(item))
                    item['expanded'] = True
                    expanded = True
                if 'fullpath' not in item:
                    if 'path' in item['parentReference']:
                        parent = re.sub(
                            '/drives/[^/]+/root:/{0,1}',
                            '',
                            item['parentReference']['path']
                        )
                        name = item['name']
                        if parent:
                            name = '/'.join([parent, name])
                        fullpath = []
                        for tok in name.split('/'):
                            while len(tok.encode('utf-8')) > 255:
                                item['mangled_path'] = True
                                for i in range(0, len(tok)):
                                    if len(tok[:i].encode('utf-8')) > 255:
                                        i -= 1
                                        break
                                fullpath.append(tok[:i])
                                tok = tok[i:]
                            fullpath.append(tok)
                        item['fullpath'] = '/'.join(fullpath)
                    else:
                        item['fullpath'] = '/'

        return items

    def verify_file(self, dest, size = None, file_hash = None, strict = True):
        if not os.path.exists(dest):
           self.logger.debug(u'{} does not exist'.format(dest))
           return False

        if strict and size is None and file_hash is None:
            self.logger.debug(u'No size or hash provided for {}'.format(dest))
            return False

        if size is not None:
            stat = os.stat(dest)
            if stat.st_size != size:
                self.logger.debug(u'{} is the wrong size: expected {}, got {}'.format(dest, size, stat.st_size))
                return False

        if file_hash:
            h = quickxorhash.QuickXORHash()
            real_hash = h.hash_file(dest)
            if real_hash != file_hash:
                self.logger.debug(u'{} has the wrong hash: expected {}, got {}'.format(dest, file_hash, real_hash))
                return False

        return True

    def _download(self, url, dest, calculate_hash = False):
        destdir = os.path.dirname(dest)
        if not os.path.exists(destdir):
            os.makedirs(destdir, 0755)

        h = None
        if calculate_hash:
            h = quickxorhash.QuickXORHash()

        try:
            with self.msgraph.get(url, stream = True) as r:
                r.raise_for_status()
                if r.headers['content-type'].startswith('multipart/'):
                    decoder = requests_toolbelt.MultipartDecoder.from_response(r)
                    for part in decoder.parts:
                        with open('{}.{}'.format(dest, part.headers['content-type'].split(';')[0].replace('/', '_')), 'wb') as f:
                            f.write(part.content)
                else:
                    with open(dest, 'wb') as f:
                        for chunk in r.iter_content(chunk_size = 1024 * 1024):
                            f.write(chunk)
                            if h is not None:
                                h.update(bytearray(chunk))
        except requests.exceptions.RequestException as e:
            self.logger.warn(e)
            return None
        if calculate_hash:
            return h.finalize()
        return True

    def download_file(self, drive_id, file_id, dest):
        url = self.get('drives/{}/items/{}/content'.format(drive_id, file_id))

        if url:
            return self._download(url['location'], dest, True)
        else:
            self.logger.warn('Failed to fetch download link from API')
            return None

    def list_notebooks(self, user):
        notebooks = self.get('users/{}@{}/onenote/notebooks?expand=sections'.format(user, self.config['domain']))
        if notebooks:
            for n in notebooks['value']:
                for s in n['sections']:
                    s['pages'] = self.get(s['pagesUrl'])['value']
            return notebooks['value']
        return []

    def _convert_page(self, page_url, page_name, dest, quirky):
        if not os.path.exists(dest + '/data'):
            os.makedirs(dest + '/data', 0755)
        raw_path = '/'.join([dest, 'raw', page_name, 'api_response'])
        result = self._download(page_url, raw_path)
        ink_file = '.'.join([raw_path, 'application_inkml+xml'])
        converter = inkml.InkML(ink_file)
        svg_file = '{}/data/{}.ink.{}'.format(dest, page_name, '{}.svg')
        converter.save(svg_file, quirky)
        raw_file = '.'.join([raw_path, 'text_html'])
        with open(raw_file, 'rb') as f:
            html = BeautifulSoup(f, 'lxml')

        # Download images and update references
        for img in html.find_all('img'):
            img_id = img['data-fullres-src'].split('/')[7].split('!')[0]
            img_file = '{}/data/{}.{}'.format(dest, img_id, img['data-fullres-src-type'].split('/')[1])
            img['src'] = 'data/' + os.path.basename(img_file)
            self._download(img['data-fullres-src'], img_file)
            for cruft in ('data-fullres-src', 'data-fullres-src-type', 'data-src-type'):
                if img.get(cruft):
                    del img[cruft]

        # Download objects and turn them into links
        for obj in html.find_all('object'):
            obj_id = obj['data'].split('/')[7]
            obj_file = '{}/data/{}.{}'.format(dest, obj_id, obj['data-attachment'])
            link = html.new_tag(
                'a',
                href = 'data/' + os.path.basename(obj_file),
                download = obj['data-attachment']
            )
            link.append(obj['data-attachment'])
            self._download(obj['data'], obj_file)
            obj.replace_with(link)

        # Check for failed export. Notably, mathematical expressions don't work.
        unexported = False
        for div in html.find_all('div'):
            if ' Processing ParagraphNode failed ' in div.contents:
                unexported = True
                if quirky:
                    # Add a visual indicator of missing data
                    img = html.new_tag('img', src = 'data/ketsuban.png')
                    div.append(img)

        if unexported:
            self.logger.warn(u'{} contained unexportable data'.format(dest))
            with open(dest + '/data/ketsuban.png', 'wb') as f:
                f.write(base64.b64decode(KETSUBAN))

        # Add InkML SVG, if it was generated
        for ink in converter.traces:
            div = html.new_tag('div', style = "position:absolute;left:0px;top:0px;pointer-events:none")
            replaced = False
            if quirky:
                # OneNote Online renders ink on top of other contents, not at
                # its normal Z location. I like this code so I'm leaving it in,
                # but disabling it by default.
                for child in html.body.children:
                    if child == ' InkNode is not supported ' and not replaced:
                        child.replace_with(div)
                        replaced = True
            if not replaced:
                html.body.append(div)
            img = html.new_tag(
                'img',
                src = 'data/' + os.path.basename(ink),
                height = '{}px'.format(converter.pixel_dimensions['Y'])
            )
            div.append(img)

        with open('{}/{}.html'.format(dest, page_name), 'wb') as f:
            f.write(html.prettify(formatter = 'html').encode('utf-8'))
        return result

    def convert_notebook(self, metadata, destdir, quirky = False):
        # quirk mode is less faithful to the official rendering, but more
        # amusing to me
        html = BeautifulSoup('<html><head></head><body></body></html>', 'lxml')
        title = html.new_tag('title')
        title.string = metadata['displayName']
        html.head.append(title)

        basedir = '/'.join([destdir, metadata['displayName']])
        if not os.path.exists(basedir):
            os.makedirs(basedir, 0755)

        for section in metadata['sections']:
            div = html.new_tag('div')
            html.body.append(div)
            heading = html.new_tag('h2')
            heading.string = section['displayName']
            div.append(heading)
            page_list = html.new_tag('ul')
            div.append(page_list)

            for page in section['pages']:
                self._convert_page(
                    page['contentUrl'] + '?includeInkML=true',
                    page['id'],
                    basedir,
                    quirky,
                )

                link = html.new_tag('a', href = page['id'] + '.html')
                link.string = page['title'] if page['title'] else 'Untitled Page'
                li = html.new_tag('li')
                li.append(link)
                page_list.append(li)

        with open('/'.join([destdir, metadata['displayName'], 'index.html']), 'wb') as f:
            f.write(html.prettify(formatter = 'html').encode('utf-8'))



#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json
import sys

from requests.exceptions import HTTPError

import odm.cli

def main():
    odm.cli.CLI.writer_wrap(sys)
    cli = odm.cli.CLI(['user', 'action', '--incremental'])
    client = cli.client

    if cli.args.action == 'show':
        user = client.show_user(cli.args.user)
        if user:
            print(json.dumps(client.show_user(cli.args.user), indent = 2))
        else:
            print('User {} not found'.format(cli.args.user), file = sys.stderr)
            sys.exit(1)

    elif cli.args.action == 'list-drives':
        drives = client.list_drives(cli.args.user)
        print(json.dumps(drives, indent = 2))

    elif cli.args.action == 'list-items':
        if not client.show_user(cli.args.user):
            print('User {} not found'.format(cli.args.user), file = sys.stderr)
            sys.exit(1)

        drives = client.list_drives(cli.args.user)
        items = client.expand_items([drive['root'] for drive in drives])
        if cli.args.incremental:
            with open(cli.args.incremental, 'rb') as f:
                old = json.load(f)['items']
                items = [x for x in items if x not in old]
        print(json.dumps({ 'items': items }, indent = 2))

    elif cli.args.action == 'list-notebooks':
        # This consistently throws a 403 for some users
        try:
            notebooks = client.list_notebooks(cli.args.user)
        except HTTPError:
            notebooks = []
        print(json.dumps({ 'notebooks': notebooks }, indent = 2))

    else:
        print('Unsupported action {}'.format(cli.args.action), file = sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import datetime
import json
import os
import sys

import odm.cli

def main():
    odm.cli.CLI.writer_wrap(sys)
    cli = odm.cli.CLI(['path', 'action', 'user', '--dest'], 'google')
    client = cli.client
    dir_map = {}

    ts_start = datetime.datetime.now()
    retval = 0

    dest = cli.args.dest if cli.args.dest else 'Migrated from OneDrive'

    if cli.args.action in ('upload', 'verify'):
        count = 0
        size = 0

        parent = None
        for tok in dest.split('/'):
            if cli.args.action == 'upload':
                parent = client.create_file(tok, parent, folder = True)
            else:
                parent = client.find_file(tok, parent)
                parent = parent['id'] if parent else None
        dir_map['.'] = parent

        for root, dirs, files in os.walk(cli.args.path):
            parent = os.path.relpath(root, cli.args.path)
            for dname in dirs:
                relpath = os.path.relpath('/'.join((root, dname)), cli.args.path)
                cli.logger.info('Working on folder {}'.format(relpath))
                if cli.args.action == 'upload':
                    dir_map[relpath] = client.create_file(dname, dir_map[parent], folder = True)
                elif dir_map[parent]:
                    existing = client.find_file(dname, dir_map[parent])
                    dir_map[relpath] = existing['id'] if existing else None
                else:
                    dir_map[relpath] = None
            for fname in files:
                count += 1
                fpath = '/'.join((root, fname))
                stat = os.stat(fpath)
                size += stat.st_size
                relpath = os.path.relpath(fpath, cli.args.path)
                cli.logger.info('Working on file {}'.format(relpath))
                if cli.args.action == 'upload':
                    attempt = 0
                    result = False
                    while attempt < 3 and not result:
                        attempt += 1
                        result = client.upload_file(fpath, fname, dir_map[parent])
                    if not result:
                        cli.logger.warn('Failed to upload {}'.format(relpath))
                        retval = 1
                elif dir_map[parent]:
                    existing = client.verify_file(fpath, fname, dir_map[parent])
                    if existing:
                        if existing['verified']:
                            cli.logger.info('Verified {}'.format(relpath))
                        else:
                            cli.logger.warn('Failed to verify {}: digest mismatch'.format(relpath))
                            retval = 1
                    else:
                        cli.logger.warn('Failed to verify {}: not found'.format(relpath))
                        retval = 1
                else:
                    cli.logger.warn('Failed to verify {}: parent folder does not exist'.format(relpath))
                    retval = 1

        cli.logger.info('{:.2f} MiB across {} items, elapsed time {}'.format(
            size / (1024 ** 2),
            count,
            datetime.datetime.now() - ts_start,
        ))

    else:
        print('Unsupported action {}'.format(cli.args.action), file = sys.stderr)
        retval = 1

    sys.exit(retval)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import os
import sys

def main():
    if len(sys.argv) < 2:
        print('Usage: odm <command> [<args>]', file = sys.stderr)
        sys.exit(1)

    cmd = os.path.basename(sys.argv[0])
    if cmd.startswith('odm'):
        cmd = 'odm'
    elif cmd.startswith('gdm'):
        cmd = 'gdm'
    else:
        print('Failed to figure out what I am: {} is an invalid wrapper name'.format(sys.argv[0]))
        sys.exit(1)

    # This is incredibly ugly, but it works.
    exec('from odm.libexec import {}_{} as subcommand'.format(cmd, sys.argv[1]))
    sys.argv[0] += ' ' + sys.argv[1]
    del(sys.argv[1])
    subcommand.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json
import os
import random
import requests
import sys
import time

from datetime import datetime
from hashlib import md5

import google.oauth2.service_account
import google.auth.transport.requests

class GoogleDriveClient:
    def __init__(self, config, logger):
        self.baseurl = 'https://www.googleapis.com/'
        self.config = config
        self.logger = logger

        cred_kwargs = {
            'subject': '{}@{}'.format(config['args'].user, config['domain']),
            'scopes': ['https://www.googleapis.com/auth/drive'],
        }
        if isinstance(config['google']['service_credentials'], dict):
            self.creds = google.oauth2.service_account.Credentials.from_service_account_info(
                config['google']['service_credentials'],
                **cred_kwargs
            )
        else:
            self.creds = google.oauth2.service_account.Credentials.from_service_account_file(
                config['google']['service_credentials'],
                **cred_kwargs
            )
        self.session = google.auth.transport.requests.AuthorizedSession(self.creds)

    def _request(self, verb, path, **kwargs):
        if self.baseurl not in path:
            path = ''.join([self.baseurl, path])

        kwargs['timeout'] = self.config.get('timeout', 60)
        kwargs['allow_redirects'] = False
        return self.session.request(verb, path, **kwargs)

    def request(self, verb, path, **kwargs):
        result = None
        attempt = 0
        while not result:
            try:
                result = self._request(verb, path, **kwargs)
            except requests.exceptions.RequestException as e:
                self.logger.warn(e)
                continue

            if result.status_code == 403 and attempt > 3:
                result.raise_for_status()
            if result.status_code in [403, 429, 500]:
                result = None
                attempt += 1
                # Jittered backoff
                delay = random.uniform(0, min(300, 3 * 2 ** attempt))
                self.logger.warn('Throttled, sleeping for {} seconds'.format(delay))
                time.sleep(delay)
            else:
                result.raise_for_status()

        return result

    def find_item(self, name, parent = None):
        query = "name = '{}' and trashed = false".format(name.replace("'", "\\'"))
        if parent:
            query = "{} and '{}' in parents".format(query, parent)

        response = self.request(
            'GET', 'drive/v3/files',
            params = {'q': query, 'fields': 'files(id,md5Checksum,size)'},
        )
        return response

    def find_file(self, name, parent = None):
        existing = self.find_item(name, parent).json()
        if len(existing['files']) > 0:
            return existing['files'][0]
        return None

    def create_file(self, name, parent = None, folder = False, mtime = None):
        existing = self.find_file(name, parent)
        if existing:
            return existing['id']

        self.logger.debug('Creating {} under {}'.format(name, parent))
        payload = {
            'name': name,
        }

        if folder:
            payload['mimeType'] = 'application/vnd.google-apps.folder'
        if parent:
            payload['parents'] = [ parent ]
        if mtime:
            payload['modifiedTime'] = mtime

        return self.request(
            'POST', 'drive/v3/files',
            json = payload
        ).json()['id']

    def verify_file(self, file_name, name, parent):
        existing = self.find_item(name, parent).json()
        if len(existing['files']) == 0:
            return None

        h = md5()
        with open(file_name, 'rb') as f:
            while True:
                block = f.read(64 * 1024)
                if block:
                    h.update(block)
                else:
                    break
        ret = existing['files'][0]
        ret['verified'] = ret['md5Checksum'] == h.hexdigest()
        return ret

    def upload_file(self, file_name, name, parent):
        stat = os.stat(file_name)
        seek = 0
        attempt = 0
        path = None
        result = None
        mtime = datetime.fromtimestamp(stat.st_mtime).isoformat() + 'Z'

        if stat.st_size == 0:
            # Metadata-only creation
            self.create_file(name, parent, mtime = mtime)
            return True

        existing = self.verify_file(file_name, name, parent)
        if existing:
            if existing['verified']:
                self.logger.debug('Verified {}'.format(file_name))
                return True

            upload = self.request(
                'PATCH',
                'upload/drive/v3/files/{}'.format(
                    existing['id']
                ),
                params = {'uploadType': 'resumable' },
                headers = {'X-Upload-Content-Length': str(stat.st_size)},
                json = {
                    'modifiedTime': mtime,
                }
            )
        else:
            upload = self.request(
                'POST',
                'upload/drive/v3/files',
                params = {'uploadType': 'resumable'},
                headers = {
                    'X-Upload-Content-Length': str(stat.st_size)
                },
                json = {
                    'name': name,
                    'parents': [ parent ],
                    'modifiedTime': mtime,
                },
            )
        path = upload.headers['location']

        self.logger.debug('Uploading {}'.format(file_name))
        while not result:
            if attempt > 0:
                # Determine the status of the upload
                try:
                    result = self.request(
                        'PUT', path,
                        headers = {'Content-Range': 'bytes */{}'.format(stat.st_size)}
                    )
                except requests.exceptions.RequestException as e:
                    self.logger.warn(e)
                    return False
                else:
                    if result.status_code in [200, 201]:
                        return True
                    elif result.status_code == 308:
                        if 'range' in result.headers:
                            seek = result.headers['range'].split('-')[1]
                        self.logger.debug('Resuming upload at byte {}'.format(seek))
                    else:
                        self.logger.warn('Upload resumption failed, HTTP {}'.format(result.status_code))
                        return False

            attempt += 1
            with open(file_name, 'rb') as f:
                f.seek(seek)
                try:
                    result = self._request(
                        'PUT', path,
                        data = f,
                        headers = {
                            'Content-Range': 'bytes {}-{}/{}'.format(
                                seek,
                                stat.st_size - 1,
                                stat.st_size,
                            )
                        },
                    )
                except requests.exceptions.RequestException as e:
                    self.logger.warn(e)

            if result is None or result.status_code == 403 or result.status_code >= 500:
                delay = random.uniform(0, min(300, 3 * 2 ** attempt))
                if result:
                    msg = 'HTTP {}'.format(result.status_code)
                else:
                    msg = 'Error'
                self.logger.warn('{}, sleeping for {} seconds'.format( msg, delay))
                time.sleep(delay)
                result = None
            elif result.status_code == 308:
                # Incomplete
                result = None
            elif result.status_code == 404 or result.status_code == 410:
                # Expired upload attempt
                return False
            else:
                result.raise_for_status()
        return True



#!/usr/bin/env python

# This file is part of ODM and distributed under the terms of the
# MIT license. See COPYING.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import calendar
import datetime
import json
import os
import sys
import time

import dateutil.parser

import odm.cli

def main():
    odm.cli.CLI.writer_wrap(sys)
    cli = odm.cli.CLI(['--dest', '--limit', '--exclude', '--start', '--diff', 'file', 'action'])
    client = cli.client

    ts_start = datetime.datetime.now()
    retval = 0

    with open(cli.args.file, 'rb') as f:
        metadata = json.load(f)

    destdir = cli.args.dest if cli.args.dest else '/var/tmp'

    if cli.args.action == 'convert-notebooks':
        for book in metadata['notebooks']:
            client.convert_notebook(book, destdir)

    elif cli.args.action in ('download', 'download-estimate', 'list-filenames', 'verify'):
        exclude = []
        if cli.args.exclude:
            with open(cli.args.exclude, 'rb') as f:
                exclude = [e.rstrip() for e in list(f)]

        size = 0
        count = 0
        started = not cli.args.start

        for item in metadata['items']:
            if not started:
                if item['fullpath'] == cli.args.start:
                    started = True
                else:
                    cli.logger.debug(u'Start point not reached, skipping item {}'.format(item['fullpath']))
                    continue

            if item['fullpath'] in exclude:
                cli.logger.debug(u'Skipping excluded item {}'.format(item['fullpath']))
                continue

            if cli.args.limit:
                if not item['fullpath'].startswith(cli.args.limit):
                    cli.logger.debug(u'Skipping non-matching item {}'.format(item['fullpath']))
                    continue

            if 'file' not in item:
                if 'folder' not in item and 'package' not in item:
                    cli.logger.debug(u'Skipping non-file {}'.format(item['fullpath']))
                continue

            if 'malware' in item:
                cli.logger.info(u'{} is tagged as malware and cannot be downloaded'.format(item['fullpath']))
                continue

            cli.logger.debug(u'Working on {}'.format(item['fullpath']))

            dest = '/'.join([destdir, item['fullpath']])
            digest = None
            if 'hashes' in item['file']:
                digest = item['file']['hashes']['quickXorHash']

            verify_args = {
                'dest': dest,
            }
            if (not cli.args.diff) or ('size' in cli.args.diff.split(',')):
                verify_args['size'] = item['size']
            if (not cli.args.diff) or ('hash' in cli.args.diff.split(',')):
                verify_args['file_hash'] = digest


            if cli.args.action == 'download':
                verify_args['strict'] = False
                if client.verify_file(**verify_args):
                    cli.logger.info(u'Verified {}'.format(dest))
                else:
                    cli.logger.info(u'Downloading {} to {}'.format(item['fullpath'], dest))
                    attempt = 0
                    result = None
                    while attempt < 3 and result is None:
                        attempt += 1
                        result = client.download_file(
                            item['parentReference']['driveId'],
                            item['id'],
                            dest,
                        )
                        if digest and result != digest:
                            cli.logger.info(u'{} has the wrong hash, retrying'.format(dest))
                            result = None
                    if result is None:
                        cli.logger.warn(u'Failed to download {}'.format(dest))
                        retval = 1
                    else:
                        os.utime(dest, (
                            time.time(),
                            calendar.timegm(dateutil.parser.parse(
                                item['fileSystemInfo']['lastModifiedDateTime']
                            ).timetuple())
                        ))

            elif cli.args.action == 'verify' and digest:
                if client.verify_file(**verify_args):
                    cli.logger.info(u'Verified {}'.format(dest))
                else:
                    cli.logger.warn(u'Failed to verify {}'.format(dest))
                    retval = 1

            elif cli.args.action == 'list-filenames':
                print(item['fullpath'])

        if cli.args.action == 'download-estimate':
            delta_msg = 'wild guess time {!s}'.format(
                datetime.timedelta(seconds = int(count + (size / (24 * 1024 * 1024))))
            )
        else:
            delta_msg = 'elapsed time {!s}'.format(datetime.datetime.now() - ts_start)

        cli.logger.info('{:.2f} MiB across {} items, {}'.format(
            size / (1024 ** 2),
            count,
            delta_msg
        ))

    elif cli.args.action == 'clean-filetree':
        fullpaths = [x['fullpath'] for x in metadata['items']]
        for root, dirs, files in os.walk(cli.args.dest):
            relpath = os.path.relpath(root, cli.args.dest)
            for fname in files:
                relfpath = '/'.join([relpath, fname])
                if relfpath[:2] == './':
                    relfpath = relfpath[2:]
                if unicode(relfpath, 'utf-8') not in fullpaths:
                    cli.logger.debug('Removing {}'.format(relfpath))
                    fpath = '/'.join([root, fname])
                    os.unlink(fpath)

    else:
        print('Unsupported action {}'.format(cli.args.action), file = sys.stderr)
        sys.exit(1)

    sys.exit(retval)

if __name__ == 'main':
    main()



#!/usr/bin/env python

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import svgwrite

from bs4 import BeautifulSoup

class InkML:
    def __init__(self, src):
        with open(src, 'r') as f:
            ink = BeautifulSoup(f, 'xml')

        self.ink = ink

        self.traces = []

        self.brushes = {}
        for brush in ink.find_all('inkml:brush'):
            brush_dict = {}
            for prop in brush.find_all('inkml:brushProperty'):
                brush_dict[prop['name']] = prop['value']
            self.brushes[brush['xml:id']] = brush_dict

        self.dimensions = {
            'X': 0,
            'Y': 0,
        }

        self.contexts = {}
        for ctx in ink.find_all('inkml:context'):
            context = {
                'channels': []
            }
            for fmt in ctx.find_all('inkml:traceFormat'):
                if ctx['xml:id'] in self.contexts:
                    raise ValueError("unsupported inkML (multiple trace formats?)")
                for channel in fmt.find_all('inkml:channel'):
                    context['channels'].append(channel['name'])
                    if 'max' in channel:
                        if channel['name'] in self.dimensions and channel['max'] > self.dimensions[channel['name']]:
                            self.dimensions[channel['name']] = channel['max']
                    else:
                        self.dimensions[channel['name']] = int(channel['max'])

                self.contexts[ctx['xml:id']] = context

        self.pixel_dimensions = {
            'X': int(self.dimensions['X'] * 0.0377952803522),
            'Y': int(self.dimensions['Y'] * 0.0377952803522),
        }

    def _parse_channels(self, context, point):
        retval = {}
        for chan, value in zip(self.contexts[context]['channels'], point.split(' ')):
            retval[chan] = value
        return retval

    def save(self, dest, split = False):
        real_dest = dest
        canvas = None

        iteration = 0
        for trace in self.ink.find_all('inkml:trace'):
            if canvas is None:
                if '{}' in dest:
                    real_dest = dest.format(iteration)
                    self.traces.append(real_dest)
                canvas = svgwrite.Drawing(real_dest, (self.dimensions['X'], self.dimensions['Y']), profile='tiny')
            brush = self.brushes[trace['brushRef'][1:]]
            extra = {
                'stroke': brush['color'],
                'stroke_width': max(int(brush['width']), int(brush['height'])),
                'opacity': 1.0 - float(brush['transparency']),
                'fill': 'none',
            }

            if brush['tip'] == 'ellipse':
                extra['stroke_linecap'] = 'round'
                extra['stroke_linejoin'] = 'round'
            path = None
            for point in trace.text.split(', '):
                coords = self._parse_channels(trace['contextRef'][1:], point)
                if path is None:
                    path = canvas.path('M{},{}'.format(coords['X'], coords['Y']), **extra)
                path.push('L{},{}'.format(coords['X'], coords['Y']))
            if path:
                canvas.add(path)
                if split:
                    canvas.save()
                    canvas = None
            iteration += 1

        if canvas:
            canvas.save()



import os
import sys

from setuptools import setup
from subprocess import check_output

version = check_output(['bash', os.path.join(os.path.dirname(__file__), 'version.sh')])

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name = 'odm',
    version = version,
    description = 'Storage Magic',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/UMCollab/ODM',
    author = 'Ezekiel Hendrickson',
    author_email = 'ezekielh@umich.edu',
    license = 'MIT',
    packages = [
        'odm',
        'odm.libexec',
    ],
    install_requires = [
        'beautifulsoup4',
        'google-auth',
        'lxml',
        'kitchen',
        'python-dateutil',
        'pyyaml',
        'requests',
        'requests_oauthlib',
        'requests_toolbelt',
        'svgwrite',
    ],
    entry_points = {
        'console_scripts': [
            'odm=odm.libexec.wrapper:main',
            'gdm=odm.libexec.wrapper:main',
        ]
    },
    include_package_data = True,
    zip_safe = False,
)
